#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class SnapBackupMorphValue{
		public Atom atom;
		public GenerateDAZMorphsControlUI morphControl;
		public List<DAZMorph> backups;
		public List<float> values;
		
		public SnapBackupMorphValue(){
			
		}
		
		public void Backup(List<DAZMorph> morphs){
			Delete();

			backups = new List<DAZMorph>();
			values = new List<float>();
			
			for(int i = 0; i < morphs.Count; i++){
				backups.Add(morphs[i]);
				values.Add(morphs[i].morphValue);
			}
		}
		
		public void Delete(){
			if(backups == null){
				return;
			}
			if(backups.Count > 0){
				backups.Clear();
				backups = null;
				values.Clear();
				values = null;
			}			
		}
		
		public void Recover(){
			for(int i = 0; i < backups.Count; i++){
				backups[i].SetValue(values[i] + 0.001f);
			}
		}
	}
	
	public class SnapBackup{
		public Atom atom;
		public string[] keys;
		public List<ControllerBackup> backups;
		public List<FreeControllerV3> ctrls;
		public List<Rigidbody> rbs;
		public bool isRigidbody = false;
		
		public SnapBackup(){
			
		}
		
		public void Backup(Atom atom, string[] keys, bool isRigidbody = false){
			this.atom = atom;
			this.isRigidbody = isRigidbody;
			Delete();
			BackupTemp(keys);
			if(!isRigidbody){
				AllJointOn();
				RecoverStatus();
				Delete();
				BackupTemp(keys);
			}
		}
		
		public void BackupTemp(string[] keys){
			this.keys = keys;
			backups = new List<ControllerBackup>();
			ctrls = new List<FreeControllerV3>();
			rbs = new List<Rigidbody>();
			
			if(isRigidbody){
				for(int i = 0; i < keys.Length; i++){
					Rigidbody rb = Utility.findRigidBody(atom, keys[i]);
					rbs.Add(rb);
					backups.Add(new ControllerBackup());
					backups[i].backUp(rb);
				}
			}else{
				for(int i = 0; i < keys.Length; i++){
					FreeControllerV3 fc = Utility.getFreeControllerV3(atom, keys[i]);
					ctrls.Add(fc);
					backups.Add(new ControllerBackup());
					backups[i].backUp(fc);
				}
			}
		}
					
		public void AllJointOn(){
			for(int i = 0; i < keys.Length; i++){
				FreeControllerV3 fc = Utility.getFreeControllerV3(atom, keys[i]);
				Utility.switchPositionState(atom, keys, "On");
				Utility.switchRotationState(atom, keys, "On");
			}
		}
		
		public void Delete(){
			if(backups != null){
				if(backups.Count > 0){
					backups.Clear();
					ctrls.Clear();
					rbs.Clear();
					keys = null;
				}			
			}
		}
		
		public void Recover(){
			AllJointOn();
			RecoverPosition();
			RecoverRotation();
			RecoverStatus();
		}
		
		public void RecoverTemp(){
			for(int i = 0; i < keys.Length; i++){
				if(isRigidbody){
					rbs[i] = backups[i].recovery(rbs[i]);
				}else{
					ctrls[i] = backups[i].recovery(ctrls[i]);
				}
			}
		}
		
		public void RecoverPosition(){
			for(int i = 0; i < keys.Length; i++){
				ctrls[i] = backups[i].recoveryPosition(ctrls[i]);
			}
		}
		
		public void RecoverRotation(){
			for(int i = 0; i < keys.Length; i++){
				ctrls[i] = backups[i].recoveryRotation(ctrls[i]);
			}			
		}
		
		public void RecoverStatus(){
			for(int i = 0; i < keys.Length; i++){
				ctrls[i] = backups[i].recoveryStatus(ctrls[i]);
			}						
		}
	}
	
	
	
	public class ControllerBackup{
		public Vector3 v3pos;
		public Quaternion v3rot;
		public string positionState;
		public string rotationState;
		
		public ControllerBackup(){
			v3pos = new Vector3(0f, 0f, 0f);
			v3rot = new Quaternion(0f, 0f, 0f, 0f);
			positionState = "Off";
			rotationState = "Off";
		}
		
		public void backUp(FreeControllerV3 target){
			try{
				v3pos = Utility.getPosition(target);
				v3rot = Utility.getRotation(target);
				positionState = target.currentPositionState.ToString();
				rotationState = target.currentRotationState.ToString();
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.backUp:" + e);
			}
		}
		
		public void backUp(Rigidbody target){
			try{
				v3pos = Utility.getPosition(target.position);
				v3rot = Utility.getRotation(target.rotation);
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.backUp:" + e);
			}
		}
		
		public FreeControllerV3 recovery(FreeControllerV3 target){
			try{
				target.SetPositionStateFromString(positionState);
				target.SetRotationStateFromString(rotationState);
				target.transform.SetPositionAndRotation(v3pos, v3rot);
				return target;
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.recovery:" + e);
				return null;
			}
		}

		public Rigidbody recovery(Rigidbody target){
			try{
				target.position = v3pos;
				target.rotation = v3rot;
				return target;
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.recovery:" + e);
				return null;
			}
		}

		public FreeControllerV3 recoveryStatus(FreeControllerV3 target){
			try{
				target.SetPositionStateFromString(positionState);
				target.SetRotationStateFromString(rotationState);
				return target;
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.recovery:" + e);
				return null;
			}
		}

		public FreeControllerV3 recoveryPosition(FreeControllerV3 target){
			try{
				Quaternion tarRot = Utility.getRotation(target);
				target.transform.SetPositionAndRotation(v3pos, tarRot);
				target.SetPositionStateFromString(positionState);
				return target;
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.recovery:" + e);
				return null;
			}
		}

		public FreeControllerV3 recoveryRotation(FreeControllerV3 target){
			try{
				Vector3 tarPos = Utility.getPosition(target);
				target.transform.SetPositionAndRotation(tarPos, v3rot);
				target.SetRotationStateFromString(rotationState);
				return target;
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.recovery:" + e);
				return null;
			}
		}
	}
}