#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	
	public class FaceMorphManager{
		public Atom atom;

		/*
		public string[] listMorphnameBrow = {"CTY_Brow1"};
		public string[] listMorphnameEyes = {"CTY_EYES_CLOSE"};
		*/
		public string[] listMorphnameBrow = {"CTY_Brow1", "CTY_BROW2b", "CTY_BROW3", "CTY_BROW4"};
		public string[] listMorphnameEyes = {"CTY_EYES_CLOSE", "CTY_EYES_CLOSE_3QUARTER", "CTY_EYES_CLOSE_HALF", "CTY_EYES_CLOSE_QUARTER", "CTY_EYES_CLOSE2", "CTY_EYES2"};
		public string[] listMorphnameMouth = {"CTY_MOUTH", "CTY_MOUTH2", "CTY_MOUTH3"};
		
		public string[] listMorphnameBrowBL = {"Brow Squeeze", "Pain", "Sad"};
		public string[] listMorphnameEyesBL = {"AAeyeclosebrowsqueeze2", "Eyes Closed", "Eyelids Bottom Up Left", "Eyelids Top Up Left", "Pain", "Sad"};
		public string[] listMorphnameMouthBL = {"Mouth Open", "Mouth Open Wide"};

		public string[] listType = {"brow", "eyes", "mouth"};
		public Dictionary<string, float> startTime;
		public Dictionary<string, float> peakTime;
		public Dictionary<string, float> finishTime;
		public Dictionary<string, float> initialStrength;
		public Dictionary<string, float> targetStrength;
		public Dictionary<string, float> currentStrength;
		public Dictionary<string, string[]> dicListMorphs;
		public Dictionary<string, string[]> dicListBLMorphs;
		public Dictionary<string, float> BLStrength;

		public Dictionary<string, string> currentMorph;
		public Dictionary<string, string> previousMorph;
		public Dictionary<string, DAZMorph> morphMap;
		public float peakTimeOffset   = 0.2f;
		public float finishTimeOffset = 2.8f;
		public float randomOffsetMin  = 0.0f;
		public float randomOffsetMax	= 0.3f;
		public float randomStrengthOffsetMin = 0.8f;
		public float randomStrengthOffsetMax = 1.2f;
		
		public float baseMultiplier = 1.0f;
		public float baseStrength = 0.3f;
		public float stageMax = 5.0f;
		public float lastUpdateInterval = 0.6f;
		public float lastUpdateTime = 0.0f;

		public bool expressionContinuityOn = true;
		public int continuityCount = 0;
		public int continuityCountMax = 4;
		public int continuityCountMin = 2;
		public Dictionary<string, int> lastNum;
		
		public bool isExpressionActive = false;
		public Dictionary<string, bool> isExpressionActivePartial;
		protected GenerateDAZMorphsControlUI morphControl;
		
		public FaceMorphManager(Atom atom){
			this.atom = atom;
			dicListMorphs = new Dictionary<string, string[]>();
			currentMorph = new Dictionary<string, string>();
			previousMorph = new Dictionary<string, string>();
			startTime = new Dictionary<string, float>();
			peakTime = new Dictionary<string, float>();
			finishTime = new Dictionary<string, float>();
			initialStrength = new Dictionary<string, float>();
			targetStrength = new Dictionary<string, float>();
			currentStrength = new Dictionary<string, float>();
			isExpressionActivePartial = new Dictionary<string, bool>();
			morphMap = new Dictionary<string, DAZMorph>();
			lastNum = new Dictionary<string, int>();
			
			dicListBLMorphs = new Dictionary<string, string[]>();
			BLStrength = new Dictionary<string, float>();
			
			dicListMorphs.Add("brow",  listMorphnameBrow);
			dicListMorphs.Add("eyes",  listMorphnameEyes);
			dicListMorphs.Add("mouth", listMorphnameMouth);
			
			dicListBLMorphs.Add("brow",  listMorphnameBrowBL);
			dicListBLMorphs.Add("eyes",  listMorphnameEyesBL);
			dicListBLMorphs.Add("mouth", listMorphnameMouthBL);
			
			foreach(var type in listType){
				currentMorph.Add(type,"");
				previousMorph.Add(type,"");
				initialStrength.Add(type, 0f);
				targetStrength.Add(type, 0f);
				currentStrength.Add(type, 0f);
				startTime.Add(type, 0f);
				peakTime.Add(type, 0f);
				finishTime.Add(type, 0f);
				BLStrength.Add(type, 0f);
				lastNum.Add(type, -1);
			}
			setFlags(false);
			
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			morphControl = character.morphsControlUI;
			continuityCount = 0;
			
			foreach(var type in listType){
				foreach(var name in dicListMorphs[type]){
					DAZMorph morph = morphControl.GetMorphByDisplayName(name);
					if(!morphMap.ContainsKey(name)){
						morphMap.Add(name, morph);
					}
				}
			}
			
			foreach(var type in listType){
				foreach(var name in dicListBLMorphs[type]){
					DAZMorph morph = morphControl.GetMorphByDisplayName(name);
					if(morph == null){
//						Utility.debug("morph: name=" + name + " is empty.");
					}else{
						if(!morphMap.ContainsKey(name)){
							morphMap.Add(name, morph);
						}
					}
				}
			}			
		}
		
		public void SetContinuityCount(){
			if(continuityCount == 0){
				continuityCount = UnityEngine.Random.Range(continuityCountMin, continuityCountMax);
			}
		}
		
		public void pickExpression(float currentTime, int stage, bool isDuringAfterglow, bool isStrongWave = false, string stageHeat = "cool"){
			if(currentTime < lastUpdateInterval + lastUpdateTime){ return; }
			lastUpdateTime = currentTime;
			
			if(continuityCount > 0){
				continuityCount--;
				foreach(var type in listType){
					string[] list = dicListMorphs[type];
					pickExpressionManually(currentTime, type, list[lastNum[type]], stage, isDuringAfterglow, isStrongWave, stageHeat);
				}
			}else{
				SetContinuityCount();
				foreach(var type in listType){
					pickExpressionPartially(currentTime, type, stage, isDuringAfterglow, isStrongWave, stageHeat);
				}
			}
		}
		
		public void pickExpressionPartially(float currentTime, string type, int stage, bool isDuringAfterglow, bool isStrongWave = false, string stageHeat = "cool"){
			string[] list = dicListMorphs[type];
			int num = UnityEngine.Random.Range(0, list.Length);
			if(stageHeat == "veryHot" || stageHeat == "nymphomania"){
				if(type == "mouth"){
					num = (UnityEngine.Random.Range(0, 1.0f) < 0.25f) ? 2 : 1; //CTY_MOUTH3 or CTY_MOUTH2 
				}else if(type == "eyes"){
					num = (UnityEngine.Random.Range(0, 1.0f) < 0.25f) ? 0 : num; //CTY_EYES_CLOSE
				}
			}
			lastNum[type] = num;
			pickExpressionManually(currentTime, type, list[num], stage, isDuringAfterglow, isStrongWave, stageHeat);
		}
		
		public void pickExpressionManually(float currentTime, string type, string morphName, int stage, bool isDuringAfterglow, bool isStrongWave = false, string stageHeat = "cool", float duration = 0.0f){
			setMorph(type, morphName, stage, isDuringAfterglow);
			setTargetStrengthAndTime(currentTime, type, stage, isDuringAfterglow, isStrongWave, duration, stageHeat);
		}
		
		public void openMouth(float currentTime, float duration, int stage, bool isDuringAfterglow, bool isStrongWave = false, string stageHeat = "cool"){
			if(continuityCount > 0 && lastNum["mouth"] == 2){
				return;
			}
			pickExpressionManually(currentTime, "mouth", "CTY_MOUTH2", stage, isDuringAfterglow, isStrongWave, stageHeat, duration);
		}
		
		public void setTargetStrengthAndTime(float currentTime, string type, int stage, bool isDuringAfterglow, bool isStrongWave = false, float duration = 0.0f, string stageHeat = "cool"){
			isExpressionActivePartial[type] = true;
			isExpressionActive = true;
			setTargetTime(currentTime, type, stage, isDuringAfterglow, isStrongWave, duration);
			setTargetStrength(currentTime, type, stage, isDuringAfterglow, isStrongWave, stageHeat);
		}

		public void setTargetTime(float currentTime, string type, int stage, bool isDuringAfterglow, bool isStrongWave = false, float duration = 0.0f){
			float afterglowRatio = (isDuringAfterglow) ? 2.0f : 1.0f;
			startTime[type] = currentTime;
			if(duration == 0.0f){
				peakTime[type] = startTime[type] + peakTimeOffset;
				finishTime[type] = peakTime[type] + finishTimeOffset * afterglowRatio;				
			}else{
				peakTime[type] = startTime[type] + peakTimeOffset;
				finishTime[type] = startTime[type] + duration;
			}
		}

		public void setTargetStrength(float currentTime, string type, int stage, bool isDuringAfterglow, bool isStrongWave = false, string stageHeat = "cool"){
			if(stageHeat == "veryHot" || stageHeat == "nymphomania"){
				isDuringAfterglow = false;
				isStrongWave = true;
			}
			initialStrength[type] = currentStrength[type];
			float randomRatio = UnityEngine.Random.Range(randomStrengthOffsetMin, randomStrengthOffsetMax);
			targetStrength[type] = (baseStrength + (1 - baseStrength) * (float)stage/stageMax) * randomRatio;
			targetStrength[type] *= baseMultiplier * ((isDuringAfterglow) ? 0.5f : 1.0f);
			if(targetStrength[type] < currentStrength[type]){
				targetStrength[type] = currentStrength[type];
			}
			if(isStrongWave && !isDuringAfterglow){
				if(type == "brow"){
					//nop
				}else if(type == "mouth"){
					targetStrength[type] *= 2.0f;
					if(targetStrength[type] > 1.3f){
						targetStrength[type] = 1.3f;
					}
				}else if(type == "eyes"){
					targetStrength[type] *= 2.0f;
					if(targetStrength[type] > 1.0f){
						targetStrength[type] = 1.0f;
					}
				}else{
					targetStrength[type] *= 1.5f;
					if(targetStrength[type] > 1.0f){
						targetStrength[type] = 1.0f;
					}
				}
			}
		}

		public void setMorph(string type, string morphName, int stage, bool isDuringAfterglow){
			if(previousMorph[type] == currentMorph[type]){
				previousMorph[type] = "";
			}
			previousMorph[type] = currentMorph[type];
			currentMorph[type] = morphName;			
		}
		
		public void manageStrength(float currentTime){
			if(!isExpressionActive){ return; }
			flagUpdate(currentTime);
			if(!isExpressionActive){
				setFlags(false);
				clearCurrentMorph();
				return;
			}
			
			foreach(var type in listType){
				if(isExpressionActivePartial[type]){
					calculateBLStrength(currentTime, type);				
					calculateStrength(currentTime, type);
				}
			}
			foreach(var type in listType){
				if(isExpressionActivePartial[type]){
					morphMap[currentMorph[type]].SetValue(currentStrength[type]);
				}
			}
			reduceExcludedMorphs();	
		}
		
		public void calculateStrength(float currentTime, string type){
			try{
				bool isBeforePeak = (currentTime < peakTime[type]) ? true : false;
				float fromStrength = isBeforePeak ? initialStrength[type] : targetStrength[type];
				float toStrength = isBeforePeak ? targetStrength[type] : 0.0f;
				currentStrength[type] = Utility.CalculateProgress(fromStrength, toStrength, startTime[type], finishTime[type], currentTime);
				currentStrength[type] *= (1 - BLStrength[type]);
				
//				float fromRatio = isBeforePeak ? 0.0f : 1.0f;
//				float toRatio = isBeforePeak ? 1.0f : 1.0f;
//				float ratio = Utility.CalculateProgress(fromRatio, toRatio, startTime[type], finishTime[type], currentTime);
			}catch (Exception e){
				Utility.loggerE(e);
			}
		}

		public void calculateBLStrength(float currentTime, string type){
			BLStrength[type] = 0f;
//			Utility.debug("type:" + type);
			foreach(var name in dicListBLMorphs[type]){
//				Utility.debug("type:" + type + ", name:" + name);
				if(morphMap.ContainsKey(name)){
					if(BLStrength[type] < morphMap[name].morphValue){
						BLStrength[type] = morphMap[name].morphValue;
					}
				}
			}
		}
		
		public void setFlags(bool value){
			isExpressionActive = value;
			foreach(var type in listType){
				isExpressionActivePartial[type] = value;
			}
		}
		
		public void clearCurrentMorph(){
			foreach(var type in listType){
				currentMorph[type] = "";
				previousMorph[type] = "";
			}
			setDefaultValue();
		}
		
		public void reduceExcludedMorphs(){
			foreach(var type in listType){
				foreach(var name in dicListMorphs[type]){
					if(name != currentMorph[type]){
						if(morphMap[name].morphValue <= 0.000001f){
							
						}else if(morphMap[name].morphValue < 0.01f){
							morphMap[name].SetValue(0.000001f);
						}else if(morphMap[name].morphValue >= 0.01f){
							morphMap[name].SetValue(morphMap[name].morphValue * 0.8f);
						}
					}
				}
			}			
		}
		
		public void setDefaultValue(){
			foreach(var type in listType){
				foreach(var name in dicListMorphs[type]){
					morphMap[name].SetValue(0.000001f);
				}
			}
		}
		
		public void setZeroValue(){
			foreach(var type in listType){
				foreach(var name in dicListMorphs[type]){
					morphMap[name].SetValue(0.0f);
				}
			}
		}
		
		public void flagUpdate(float currentTime){
			bool result = false;
			foreach(var type in listType){
				if(currentTime < finishTime[type]){
					isExpressionActivePartial[type] = true;
					result = true;
				}else{
					isExpressionActivePartial[type] = false;					
				}
			}
			isExpressionActive = result;
		}
		
		
		public bool isExpired(float currentTime){
			bool result = true;
			foreach(var type in listType){
				if(currentTime < finishTime[type]){
					result = false;
				}
			}
			return result;
		}
		
		public void stopExpression(){
			setDefaultValue();
			setZeroValue();
			setFlags(false);
			clearCurrentMorph();
		}
		
		public void stopAllExpressionInPose(){
			try{
				List<DAZMorph> othersMorphs = new List<DAZMorph>();
				
				morphControl.onlyShowActive = true;
				morphControl.showType = GenerateDAZMorphsControlUI.ShowType.Pose;
				List<DAZMorph> morphs = morphControl.GetMorphs();
				for(int i = 0; i < morphs.Count; i++){
					if(morphs[i].resolvedRegionName.StartsWith("Pose/Head") && morphs[i].active){
						othersMorphs.Add(morphs[i]);
					}
				}
				
				for(int i = 0; i < othersMorphs.Count; i++){
					//core.logger("Set Default Value " + othersMorphs[i].resolvedDisplayName);
					othersMorphs[i].SetDefaultValue();
				}			
			}catch (Exception e){
				Utility.loggerE(e);
			}
		}
		
		public void OnDestroy(){
			
		}
		
	}
}