using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using Object = UnityEngine.Object;
using System.Text.RegularExpressions;

namespace Cotyounoyume
{
    public class AddonTakeControlItem : MonoBehaviour 
    {
		public Atom atom;
		public EBATCore core;
		private float lastTime;
		public bool autoStopOnGlobal = true;
		public bool autoStopMorphOnGlobal = true;
		public bool autoStopHeadOnGlobal = true;
		public bool autoStopBodyOnGlobal = false;
		
		protected List<string> potentialConflictsTable;
		
		protected Dictionary<string, bool> statusList;
		public Dictionary<string, string> triggersStore;
		public Dictionary<string, string> animationPatternsTriggersStore;
		public Dictionary<string, string> animationPatternAnimatedObjectStore;
		public Dictionary<string, string> pluginIDTable;
		public Dictionary<string, List<string>> pluginIDTableList;
		public Dictionary<string, JSONClass> motionAnimationControlsStore;
		public Dictionary<string, Dictionary<string, string>> timelineControllerStore;
		
		public string headCurrentRotationState = "";
		public FreeControllerV3 headControl;
		public string snapTimeline;
		
		public TimelineLink tlLink;
		
		public string[] defaultKey = {
			"[AS]Stop Morph Control by non-plugins",
			"[AS]Stop Head Control by non-plugins",
			"[AS]Stop Body Control by non-plugins",
		};

		public string[] defaultKeyPlus1 = {
			"[AS]AutoStop Other Plugins",
			"[AS]Stop Morph Control by non-plugins",
			"[AS]Stop Head Control by non-plugins",
			"[AS]Stop Body Control by non-plugins",
		};
		
		#region takeControlOfExpression
		ReceiverBackup receiverBackup;
		VariableTriggerBackup variableTriggerBackup;
		CollisionTriggerBackup collisionTriggerBackup;
		#endregion
		
		#region addon common
		public AddonTakeControlItem(EBATCore core){
			atom = core.atom;
			this.core = core;
			statusList = new Dictionary<string, bool>();
			potentialConflictsTable = new List<string>();
			
			triggersStore = new Dictionary<string, string>();
			animationPatternsTriggersStore = new Dictionary<string, string>();
			animationPatternAnimatedObjectStore = new Dictionary<string, string>();
			pluginIDTable = new Dictionary<string, string>();
			motionAnimationControlsStore = new Dictionary<string, JSONClass>();
			timelineControllerStore = new Dictionary<string, Dictionary<string, string>>();
			pluginIDTableList = new Dictionary<string, List<string>>();

			headControl = Utility.getFreeControllerV3(core.atom, "headControl");
//			tlLink = new TimelineLink(core);
		}
		
		public void OnDestroy(){
			ReturnControlOfPlugins();
			
			triggersStore.Clear();
			statusList.Clear();
			animationPatternsTriggersStore.Clear();
			animationPatternAnimatedObjectStore.Clear();
			motionAnimationControlsStore.Clear();
			pluginIDTable.Clear();			
			pluginIDTableList.Clear();			
			potentialConflictsTable.Clear();
			timelineControllerStore.Clear();
//			defaultKey.Clear();
//			defaultKeyPlus1.Clear();
		}

		public void loadSettings(){
			try{
				autoStopOnGlobal = core.settings["General"]["AutoStop OtherPlugins"].AsBool;
				potentialConflictsTable.Clear();
				makePotentialConflictsTable();
				makePluginIDTable();
				
				if(core.settingsRoot["Common"].AsObject.HasKey("PTMC On/Off")){
					foreach(var name in potentialConflictsTable){
						if(!core.settingsRoot["Common"]["PTMC On/Off"].AsObject.HasKey(name)){
							if(name.StartsWith("[AS]Stop Body")){
								core.settingsRoot["Common"]["PTMC On/Off"][name] = "false";
							}else{
								core.settingsRoot["Common"]["PTMC On/Off"][name] = "true";
							}
						}
					}
				}else{
					foreach(var name in potentialConflictsTable){
						string value = (name.StartsWith("[AS]Stop Body")) ? "false" : "true";
						core.settingsRoot["Common"]["PTMC On/Off"][name] = value;
					}
				}
				
				if(core.settingsRoot["Common"].AsObject.HasKey("PTMC Timing")){
					foreach(var name in potentialConflictsTable){
						if(!core.settingsRoot["Common"]["PTMC Timing"].AsObject.HasKey(name)){
							core.settingsRoot["Common"]["PTMC Timing"][name] = "5";
						}
					}
				}else{
					foreach(var name in potentialConflictsTable){
						core.settingsRoot["Common"]["PTMC Timing"][name] = "5";
					}
				}
				
				if(core.settingsRoot["Common"].AsObject.HasKey("AutoStop On/Off")){
					foreach(var name in defaultKeyPlus1){
						if(!core.settingsRoot["Common"]["AutoStop On/Off"].AsObject.HasKey(name)){
							if(name.StartsWith("[AS]Stop Body")){
								core.settingsRoot["Common"]["AutoStop On/Off"][name] = "false";
							}else{
								core.settingsRoot["Common"]["AutoStop On/Off"][name] = "true";
							}
						}
					}
				}else{
					foreach(var name in defaultKeyPlus1){
						string value = (name.StartsWith("[AS]Stop Body")) ? "false" : "true";
						core.settingsRoot["Common"]["AutoStop On/Off"][name] = value;
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void addPotentialConflictsTableIndividual(string pluginName){
			if(!core.settingsRoot["Common"]["Plugins that might Confilict"].AsObject.HasKey(pluginName)){
				potentialConflictsTable.Add(pluginName);
			}
			if(!core.settingsRoot["Common"]["PTMC On/Off"].AsObject.HasKey(pluginName)){
				core.settingsRoot["Common"]["PTMC On/Off"][name] = "true";
			}
			if(!core.settingsRoot["Common"]["PTMC Timing"].AsObject.HasKey(pluginName)){
				core.settingsRoot["Common"]["PTMC Timing"][name] = "5";
			}
		}
		
		public void makePluginIDTable(){
			pluginIDTable.Clear();
			pluginIDTableList.Clear();
			List<string> ids = atom.GetStorableIDs();
			foreach(var id in ids){
				if(id.StartsWith("plugin#")){
					string pluginName = NormalizePluginName(id);
					if(!pluginIDTable.ContainsKey(pluginName)){
						pluginIDTable.Add(pluginName, id);
					}
					if(!pluginIDTableList.ContainsKey(pluginName)){
						pluginIDTableList.Add(pluginName, new List<string>());
					}
					pluginIDTableList[pluginName].Add(id);
				}
			}
		}
		
		public void makePotentialConflictsTable(){
			Dictionary<string, bool> hashMap = new Dictionary<string, bool>();
			foreach(var key in defaultKeyPlus1){
				if(!hashMap.ContainsKey(key)){
					hashMap.Add(key, true);
					potentialConflictsTable.Add(key);
					core.settingsRoot["Common"]["Plugins that might Confilict"].Add(key);
				}
			}
			
			foreach(var name in core.settingsRoot["Common"]["Plugins that might Confilict"].AsArray){
				string key = name.ToString().Trim('"');
				if(!hashMap.ContainsKey(key)){
					hashMap.Add(key, true);
					potentialConflictsTable.Add(key);
				}
			}
		}
		
		public void applySettings(){
		}

		public void manage(float time){
		}	
		
		public void OnOrgasmStart()
		{
		}
		
		public void OnOrgasmStop()
		{
		}
		#endregion
		
		#region TakeControl		
		public List<string> GetPluginList(string mode = "unlisted", bool distinctEnabled = true){
			List<string> result = new List<string>();
			List<string> ids = atom.GetStorableIDs();
			Dictionary<string, bool> dic = new Dictionary<string, bool>();
			foreach(var id in ids){
				if(id.StartsWith("plugin#")){
					string pluginName = NormalizePluginName(id);
					if(mode == "unlisted" && !isPluginListed(pluginName)){
						if(!dic.ContainsKey(pluginName)){
							dic.Add(pluginName, true);
						}
					}else if(mode == "all"){
						if(!dic.ContainsKey(pluginName)){
							dic.Add(pluginName, true);
						}
					}
				}
			}

			foreach(KeyValuePair<string, bool> kvp in dic){
				result.Add(kvp.Key);
			}
			result.Sort();
			
			return result;
		}
		
		public string NormalizePluginName(string orig){
			string[] pattern = {"_"};
			string[] words = orig.Split(pattern, StringSplitOptions.None);
			string pluginPrefix = words[0];
			string pluginName = orig.Remove(0, pluginPrefix.Length+1);
			return pluginName;
		}
		
		public void AutoStopPlugins(){
			if(pluginIDTable == null){
				makePluginIDTable();
			}
			List<string> loadedPluginList = GetPluginList("all");
			int currentStage = core.getStage();
			
			string ngWord1 = "Cotyounoyume.EBATCore";
			string ngWord2 = core.vammoanLinkOnGlobal ? "VAMMoanPlugin.VAMMoan" : "";
			foreach(var loadedPlugin in loadedPluginList){
				if(loadedPlugin == ngWord1){ continue; };
				if(loadedPlugin == ngWord2){ continue; };
				
				if(isPluginWithPotentialConflicts(loadedPlugin)){
					if(!pluginIDTableList.ContainsKey(loadedPlugin)){
						makePluginIDTable();
					}
					if(isStopTiming(loadedPlugin, currentStage)){
						if(!pluginIDTable.ContainsKey(loadedPlugin)){
							core.debug("ERROR: Unknown Bug1: " + loadedPlugin);
						}else{
							foreach(var pluginName in pluginIDTableList[loadedPlugin]){
								TemporarilyDeactivatePlugin(pluginName);
							}
						}
					}else{
						foreach(var pluginName in pluginIDTableList[loadedPlugin]){
							ReactivatePlugin(pluginName);
						}
					}
				}
			}

			foreach(var name in defaultKey){
				if(core.settingsRoot["Common"]["AutoStop On/Off"][name].AsBool){
					if(currentStage >= core.settingsRoot["Common"]["PTMC Timing"][name].AsInt){
						TakeControlOfTarget(name);
					}else{
						ReturnControlOfTarget(name);
					}
				}
			}			
		}
		
		public void TakeControlOfPlugins(){
			List<string> ids = atom.GetStorableIDs();
			int stage = core.getStage();
			foreach(var id in ids){
				if(id.StartsWith("plugin#")){
					string pluginName = NormalizePluginName(id);
					if(isPluginWithPotentialConflicts(pluginName)){
						TemporarilyDeactivatePlugin(id);
					}
				}
			}
		}
		
		public bool isStopTiming(string name, int currentStage){
			bool result = false;
			int stageTiming = core.settingsRoot["Common"]["PTMC Timing"][name].AsInt;
			if(currentStage >= stageTiming){
				result = true;
			}else{
				result = false;
			}
			return result;
		}
		
		public bool isPluginListed(string id){
			if(core.settingsRoot["Common"]["PTMC On/Off"].AsObject.HasKey(id)){
				return true;
			}else{
				return false;
			}
		}
		
		public bool isPluginWithPotentialConflicts(string id){
			if(isPluginListed(id) && core.settingsRoot["Common"]["PTMC On/Off"][id].AsBool){
				return true;
			}else{
				return false;
			}
		}

		public void TemporarilyDeactivatePlugin(string id){
			JSONStorable jss = atom.GetStorableByID(id);
			bool currentStatus = jss.GetBoolParamValue("enabled");
			if(statusList.ContainsKey(id)){
//				statusList[id] = currentStatus;
			}else{
				statusList.Add(id, currentStatus);				
			}
			jss.SetBoolParamValue("enabled", false);
		}
		
		public void ReturnControlOfPlugins(){
			foreach(KeyValuePair<string, bool> kvp in statusList){
				ReactivatePlugin(kvp.Key);
			}
			statusList.Clear();
		}
		
		public void ReactivatePlugin(string id){
			JSONStorable jss = atom.GetStorableByID(id);
			if(statusList.ContainsKey(id)){
				jss.SetBoolParamValue("enabled", statusList[id]);
			}
		}
		#endregion
		
		#region TakeControl
		public void TakeControlOfTarget(string name, bool isMorph = true, bool isSpecificControl = true){
			switch(name){
				case "[AS]Stop Morph Control by non-plugins":
					StopVariableTrigger(bMorph:isMorph);
					StopAnimationPattern(bMorph:isMorph);
					core.expression2.stopAllExpressionInPose();
					break;
				case "[AS]Stop Head Control by non-plugins":
					StopAnimatedObject(bSpecificControl:true);
					StopMotionAnimation(bSpecificControl:true);
					break;
				case "[AS]Stop Body Control by non-plugins":
					StopAnimatedObject(bSpecificControl:false);
					StopMotionAnimation(bSpecificControl:false);
					break;
			}
		}		
		
		public void TakeControlOfEverythingButPlugins(bool isMorph = true, bool isSpecificControl = true){
			StopAnimationPattern(bMorph:isMorph);
			StopVariableTrigger(bMorph:isMorph);
			StopAnimatedObject(bSpecificControl:isSpecificControl);
			StopMotionAnimation(bSpecificControl:isSpecificControl);
			StopTimeline(bSpecificControl:isSpecificControl);
		}
		
		public void ReturnControlOfTarget(string name){
			switch(name){
				case "[AS]Stop Morph Control by non-plugins":
					RestartVariableTrigger();
					RestartAnimationPattern();
					break;
				case "[AS]Stop Head Control by non-plugins":
					RestartAnimatedObject();
					RestartMotionAnimation();
					break;
				case "[AS]Stop Body Control by non-plugins":
					RestartAnimatedObject();
					RestartMotionAnimation();
					break;				
			}
		}
		
		public void ReturnControlOfEverythingButPlugins(){
			RestartAnimationPattern();
			RestartVariableTrigger();
			RestartAnimatedObject();
			RestartMotionAnimation();
			RestartTimeline();
		}
		
		public void StopAnimationPattern(bool bMorph = true){
//			core.debug("StopAnimationPattern");
			foreach(KeyValuePair<string, AnimationPattern> kvp in Utility.CollectAnimationPatterns()){
				if(Utility.ContainsAtomInTriggers(core.atom, kvp.Value, isMorph:bMorph)){
					animationPatternsTriggersStore.Add(kvp.Key, kvp.Value.GetJSON().ToString());
					JSONClass erasedJsc = Utility.MakeJSCWithAtomRelatedTriggersExcluded(core.atom, kvp.Value, isMorph:bMorph);
					kvp.Value.LateRestoreFromJSON(erasedJsc);
				}
			}
		}
					
		public void RestartAnimationPattern(){
			foreach(KeyValuePair<string, AnimationPattern> kvp in Utility.CollectAnimationPatterns()){
				string atomUid = kvp.Key;
				if(animationPatternsTriggersStore.ContainsKey(atomUid)){					
					JSONClass jsc = JSONNode.Parse(animationPatternsTriggersStore[atomUid]).AsObject;
					kvp.Value.LateRestoreFromJSON(jsc);
					animationPatternsTriggersStore.Remove(atomUid);
				}
			}			
		}
					
		public void StopAnimatedObject(bool bSpecificControl = true){
//			core.debug("StopAnimatedObject");
			foreach(KeyValuePair<string, JSONStorable> kvp in Utility.CollectAnimatedObjects()){
				if(Utility.ContainsAtomInAnimationTarget(core.atom, kvp.Value, isSpecificControl:bSpecificControl)){
					animationPatternAnimatedObjectStore.Add(kvp.Key, kvp.Value.GetJSON().ToString());
					JSONClass erasedJsc = Utility.MakeJSCWithAtomRelatedReceiverExcluded(core.atom, kvp.Value, isSpecificControl:bSpecificControl);
					kvp.Value.RestoreFromJSON(erasedJsc);
				}
			}
		}
				
		public void RestartAnimatedObject(){
			foreach(KeyValuePair<string, JSONStorable> kvp in Utility.CollectAnimatedObjects()){
				string atomUid = kvp.Key;
				if(animationPatternAnimatedObjectStore.ContainsKey(atomUid)){
					JSONClass jsc = JSONNode.Parse(animationPatternAnimatedObjectStore[atomUid]).AsObject;
					kvp.Value.RestoreFromJSON(jsc);
					animationPatternAnimatedObjectStore.Remove(atomUid);
				}
			}
		}

		public void StopVariableTrigger(bool bMorph = true){
//			core.debug("StopVariableTrigger");
			foreach(KeyValuePair<string, VariableTrigger> kvp in Utility.CollectVariableTriggers()){
				if(Utility.ContainsAtomInTrigger(core.atom, kvp.Value, isMorph:bMorph)){
					triggersStore.Add(kvp.Key, kvp.Value.GetJSON().ToString());
					JSONClass erasedJsc = Utility.MakeJSCWithAtomRelatedTriggerExcluded(core.atom, kvp.Value, isMorph:bMorph);
					kvp.Value.LateRestoreFromJSON(erasedJsc);
				}
			}
		}
		
		public void RestartVariableTrigger(){
			foreach(KeyValuePair<string, VariableTrigger> kvp in Utility.CollectVariableTriggers()){
				string atomUid = kvp.Key;
				if(triggersStore.ContainsKey(atomUid)){
					JSONClass jsc = JSONNode.Parse(triggersStore[atomUid]).AsObject;
					kvp.Value.LateRestoreFromJSON(jsc);
					triggersStore.Remove(atomUid);
				}
			}
		}
		
		public void StopMotionAnimation(bool bSpecificControl = true){
//			core.debug("StopMotionAnimation");
			MotionAnimationControl[] MAC = core.atom.motionAnimationControls;
			foreach(var mac in MAC){
				JSONClass jsc = mac.GetJSON();
				string id = jsc["id"].Value.Trim('"');
				if(!bSpecificControl || Utility.CheckSpecificControl(id)){
					if(!motionAnimationControlsStore.ContainsKey(id)){
						motionAnimationControlsStore.Add(id, jsc);
					}else{
						motionAnimationControlsStore[id] = jsc;
					}
					mac.ClearAnimation();
				}
			}
		}
		
		public void RestartMotionAnimation(){
			MotionAnimationControl[] MAC = core.atom.motionAnimationControls;
			for(int i = 0; i < MAC.Length; i++){
				JSONClass jsc = MAC[i].GetJSON();
				string id = jsc["id"].Value.Trim('"');
				if(motionAnimationControlsStore.ContainsKey(id)){
					MAC[i].RestoreFromJSON(motionAnimationControlsStore[id]);
				}
			}
			motionAnimationControlsStore.Clear();
		}
		
		public string GetTimelineName(){
			string timelineName = "";
			List<string> names = Utility.GetPluginNamesFromAtom(core.atom);
			foreach(var name in names){
				if(name.Contains("Timeline")){
					timelineName = name;
				}
			}
			return timelineName;
		}
		
		
		public void StopTimeline(bool bMorph = true, bool bSpecificControl = true){
			
//			core.isUnderControl = true;
//			headControl.SetLinkToAtom(core.trunkHead.uid);
//			headCurrentRotationState = headControl.currentRotationState.ToString();
//			headControl.SetRotationStateFromString("PhysicsLink");
		}
		
		public void RestartTimeline(){
//			headControl.SetLinkToAtom("");
//			headControl.SetRotationStateFromString(headCurrentRotationState);
//			core.isUnderControl = false;
		}
		#endregion
		
		public void CollectTimelineExpandClickable(){
			core.StartCoroutine(CollectTimelineExpandClickableCoroutine());
		}
		
		public IEnumerator CollectTimelineExpandClickableCoroutine(){
//			if(tlLink.IsInitted()){
			if(false){
				if(tlLink.HasPositionAndRotation("head")){
					tlLink.ChangePositionAndRotationDirty("head", "Toggle");
					core.debug("HasPositionAndRotation OK");				
				}else{
					core.debug("HasPositionAndRotation NG");				
				}
				yield return null;
			}else{
				Utility.OpenAtomUI(core.atom);
				if(!Utility.OpenCustomUI(core.atom, "Timeline.AtomPlugin")){ yield break; }

				yield return new WaitForSeconds(0.1f);
	//			yield return null;
				
				tlLink.Init();

				yield return new WaitForSeconds(0.1f);
	//			yield return null;
				
				tlLink.SetState("head", true);
				yield return new WaitForSeconds(0.1f);
	//			yield return null;
				
				tlLink.SetPositionAndRotationToggle("head");
				tlLink.SetPositionAndRotationUIs("head");
				yield return new WaitForSeconds(0.5f);
	//			yield return null;
				
				Utility.CloseUI();			
			}
		}
		
		
	}
	
	public class TimelineLink{
		public Atom atom;
		public EBATCore core;
		public Transform rootT;
		public Dictionary<string, TimelineToggle> toggleMap;
		public bool initted = false;
		
		public TimelineLink(EBATCore core){
			atom = core.atom;
			this.core = core;
			
			rootT = Utility.GetTimelineScriptUIRootTransform(atom);
		}
		
		public void Init(){
			if(IsInitted()){ return;}
			
			toggleMap = new Dictionary<string, TimelineToggle>();
			foreach(var t in rootT.GetComponentsInChildren<UIDynamicToggle>()){
				if(t.name == "DynamicToggle(Clone)" && t.transform.parent.name == "DynamicSpacer(Clone)"){
					string text = t.GetComponentInChildren<Text>().text;
					if(!toggleMap.ContainsKey(t.name)){
						toggleMap.Add(text, new TimelineToggle(text, t.transform.parent));
					}else{
						toggleMap[text] = new TimelineToggle(text, t.transform.parent);
					}
				}
			}
			initted = true;
		}
		
		public bool IsInitted(){
			return initted;
		}
		
		public bool ContainsKey(string name){
			return toggleMap.ContainsKey(name);
		}
		
		public bool IsOpen(string name){
			if(!ContainsKey(name)){ return false; }
			return toggleMap[name].GetState();
		}
		
		public void Toggle(string name){
			if(!ContainsKey(name)){ return; }
			toggleMap[name].ToggleState();
		}
		
		public void SetPositionAndRotationUIs(string name){
			if(!ContainsKey(name)){ return; }
			
			UIDynamicToggle tmpPosition = null;
			UIDynamicToggle tmpRotation = null;
			
			if(IsOpen(name)){
				foreach(var t in rootT.GetComponentsInChildren<UIDynamicToggle>()){
					string text = t.GetComponentInChildren<Text>().text;
					if(t.name == "DynamicToggle(Clone)" && text == "Position"){
						tmpPosition = t;
					}else if(t.name == "DynamicToggle(Clone)" && text == "Rotation"){
						tmpRotation = t;
					}
				}
				
				if(tmpPosition != null && tmpRotation != null){
					toggleMap[name].SetPositionAndRotationUIs(tmpPosition, tmpRotation);
				}
			}
		}
		
		public bool HasPositionAndRotation(string name){
			if(!ContainsKey(name)){ return false; }
			return toggleMap[name].HasPositionAndRotation();			
		}
		
		public void ChangePositionAndRotationDirty(string name, string type){
			if(!ContainsKey(name)){ return; }
			toggleMap[name].ChangePositionAndRotationDirty(type);			
		}
		
		public void SetPositionAndRotationOff(string name){
			if(!ContainsKey(name)){ return; }
			if(IsOpen(name)){
				foreach(var t in rootT.GetComponentsInChildren<UIDynamicToggle>()){
					string text = t.GetComponentInChildren<Text>().text;
					if(t.name == "DynamicToggle(Clone)" && (text == "Position" || text == "Rotation")){
						t.toggle.isOn = false;
					}
				}
			}
		}
		
		public void SetPositionAndRotationOn(string name){
			if(!ContainsKey(name)){ return; }
			if(IsOpen(name)){
				foreach(var t in rootT.GetComponentsInChildren<UIDynamicToggle>()){
					string text = t.GetComponentInChildren<Text>().text;
					if(t.name == "DynamicToggle(Clone)" && (text == "Position" || text == "Rotation")){
						t.toggle.isOn = true;
					}
				}
			}
		}
		
		public void SetPositionAndRotationToggle(string name){
			if(!ContainsKey(name)){ return; }
			if(IsOpen(name)){
				foreach(var t in rootT.GetComponentsInChildren<UIDynamicToggle>()){
					string text = t.GetComponentInChildren<Text>().text;
					if(t.name == "DynamicToggle(Clone)" && (text == "Position" || text == "Rotation")){
						t.toggle.isOn = !t.toggle.isOn;
					}
				}
			}
		}
		
		public void SetState(string name, bool value){
			if(!ContainsKey(name)){ return; }
			toggleMap[name].SetState(value);
		}
		
		public bool IsValid(){
			bool result = true;
			foreach(KeyValuePair<string, TimelineToggle> kvp in toggleMap){
				if(toggleMap[kvp.Key].IsValid() == false){
					result = false;
				}
			}
			return result;
		}
		
	}
	
	public class TimelineToggle{
		public string name;
		public Transform expander;
		public bool isOpen;
		public bool bValid;
		public UIDynamicToggle togglePosition;
		public UIDynamicToggle toggleRotation;

		public TimelineToggle(string name, Transform t){
			this.name = name;
			this.expander = t;			
			this.isOpen = false;
			bValid = false;
			SyncState();
		}
		
		public void SyncState(){
			foreach(Transform t in expander){
				Text textObj = t.GetComponentInChildren<Text>();
				string text = (textObj == null) ? "" : textObj.text;
				isOpen = (text == "˅") ? true : isOpen;
				bValid = true;
			}
		}
		
		public bool HasPositionAndRotation(){
			return (togglePosition != null && toggleRotation != null);
		}
		
		public void ChangePositionAndRotationDirty(string type){
			if(HasPositionAndRotation()){
				if(type == "On"){
					togglePosition.toggle.isOn = true;
					toggleRotation.toggle.isOn = true;
				}else if(type == "Off"){
					togglePosition.toggle.isOn = false;
					toggleRotation.toggle.isOn = false;
				}else if(type == "Toggle"){
					togglePosition.toggle.isOn = !togglePosition.toggle.isOn;
					toggleRotation.toggle.isOn = !toggleRotation.toggle.isOn;
				}
			}
		}
		
		public void SetPositionAndRotationUIs(UIDynamicToggle togglePosition, UIDynamicToggle toggleRotation){
			this.togglePosition = togglePosition;
			this.toggleRotation = toggleRotation;
		}
		
		public bool GetState(){
			if(expander != null){
				return isOpen;
			}else{
				bValid = false;
				return false;
			}
		}
		
		public void SetState(bool value){
			if(value != GetState()){
				ToggleState();				
			}
		}
		
		public void ToggleState(){
			if(expander != null){
				expander.SendMessage("ToggleExpanded");
				SyncState();
			}
		}
		
		public bool IsValid(){
			return bValid;
		}
	}
}