#region using
using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
using MeshVR;
#endregion

namespace Cotyounoyume{
	public class ClothesStateChanger
	{
		List<ClothStateChanger> clothStateChangerArray = new List<ClothStateChanger>();
		DAZClothingItem test;
		Atom atom;
		DAZSkinWrap skinwrap1;
		List<DAZClothingItem> myClothes;
		EBATCore core;
		
		public ClothesStateChanger(Atom atom, EBATCore core){
			this.core = core;
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			
			myClothes = new List<DAZClothingItem>();
			this.atom = atom;
			ChangeClothes();
		}
		
		public void ChangeClothes(){
			try{
				myClothes.Clear();
				clothStateChangerArray.Clear();
				
				JSONStorable geometryStorable = atom.GetStorableByID("geometry");
				DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
				
				myClothes = new List<DAZClothingItem>();
				clothStateChangerArray = new List<ClothStateChanger>();
				
				List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
				for (int i = 0; i < clothes.Count; i++){
					DAZClothingItem c = clothes[i];
					if(c.active && c.creatorName != "cotyounoyume"){
						clothStateChangerArray.Add(new ClothStateChanger(c));
						myClothes.Add(c);
					}
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void RemoveClothes(){
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			foreach(var cloth in myClothes){
				geometryForClothes.SetActiveClothingItem(cloth, active: false);
			}
		}
		
		public void RemoveResetClothes(){
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			foreach(var cloth in myClothes){
				geometryForClothes.SetActiveClothingItem(cloth, active: true);
			}
		}
		
		public void CheckClothAttribute(){
			ChangeClothes();
			for(int i = 0; i < clothStateChangerArray.Count; i++){
				clothStateChangerArray[i].CheckClothAttribute();
			}
			
		}
		
		public void Recover(){
			try{
				for(int i = 0; i < clothStateChangerArray.Count; i++){
					clothStateChangerArray[i].Recover();
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void Reset(){
			try{
				for(int i = 0; i < clothStateChangerArray.Count; i++){
					clothStateChangerArray[i].Reset();
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void WrapAlter(bool isForce = false){
			for(int i = 0; i < clothStateChangerArray.Count; i++){
				clothStateChangerArray[i].WrapAlter();
			}
		}
		
		public void Change(string type = "UndressStrong", float ratio = 1.0f){
			try{
				for(int i = 0; i < clothStateChangerArray.Count; i++){
					clothStateChangerArray[i].Change(type, ratio);
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void UndressStop(){
			try{
				for(int i = 0; i < clothStateChangerArray.Count; i++){
					clothStateChangerArray[i].UndressStop();
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void OnDestroy(){
			for(int i = 0; i < clothStateChangerArray.Count; i++){
				clothStateChangerArray[i].OnDestroy();
				clothStateChangerArray[i] = null;
			}
			clothStateChangerArray = null;
		}
	}
	
	public class ClothStateChanger
	{
		public List<ClothSimStateChanger> cssc = new List<ClothSimStateChanger>();
		private DAZClothingItem cloth;
		public bool hasAltWrap = false;
		public List<string> AltWrapNames;
		public DAZSkinWrapSwitcher wrapSwitcher = null;
		public bool simEnabledOrig = false;
		public bool isBuiltin = false;
		public bool hasRuntime = false;
		public bool hasCsc = false;
		public bool hasClothSettings = false;
		
		public ClothStateChanger(DAZClothingItem cloth){
			this.cloth = cloth;
			AltWrapNames = new List<string>();
			
			CollectClothInfo();
		}

		public void CollectClothInfo(){
			isBuiltin = !cloth.uid.Contains("/");
			ClothSimControl[] cscArray = this.cloth.GetComponentsInChildren<ClothSimControl>();
			foreach (ClothSimControl csc in cscArray){
				var css = csc.clothSettings;
				if(css != null){
					hasRuntime = (css.Runtime != null) ? true : false;
					hasClothSettings = true;
				}
				hasCsc = true;
				simEnabledOrig = (csc.simEnabledJSON.val) ? true : simEnabledOrig;
				cssc.Add(new ClothSimStateChanger(csc));
			}
			
			wrapSwitcher = this.cloth.GetComponentInChildren<DAZSkinWrapSwitcher>();
			if(wrapSwitcher != null && wrapSwitcher.currentWrapName != ""){
				hasAltWrap = true;
				foreach(var name in wrapSwitcher.GetStringChooserJSONParamChoices("wrapName")){
					if(name != "Normal"){
						AltWrapNames.Add(name);
					}
				}
			}else{
				wrapSwitcher = null;
			}
		}
		
		public void CheckClothAttribute(){
			Utility.debug("clothname:" + cloth.displayName);
			Utility.debug("  uid:" +  cloth.uid);
			Utility.debug("  isBuiltin:" + isBuiltin);
			Utility.debug("  hasRuntime:" + hasRuntime);
			Utility.debug("  hasCsc:" + hasCsc);
			Utility.debug("  simEnabledOrig:" + simEnabledOrig);
			Utility.debug("  hasAltWrap:" + hasAltWrap);
//			Utility.debug("  exclusiveRegion:" + cloth.exclusiveRegion); //always None...
//			Utility.debug("  drawRigidOnBoneType:" + cloth.drawRigidOnBoneType); //always None...
//			Utility.debug("  drawRigidOnBone:" + cloth.drawRigidOnBone); //always Null...
			Utility.debug("  skin:" + cloth.skin); //always None...
			Utility.debug("  isDynamicRuntimeLoaded:" + cloth.isDynamicRuntimeLoaded);
			Utility.debug("  dynamicRuntimeLoadPath:" + cloth.dynamicRuntimeLoadPath);
			Utility.debug("  instanceName:" + cloth.instanceName);
			var dazMeshes = cloth.GetComponentsInChildren<DAZMesh>();
			foreach(var dazMesh in dazMeshes){
				Utility.debug("    dazMesh:" + dazMesh);
				Utility.debug("    baseVertices:" + dazMesh.baseVertices);
				Utility.debug("    baseVertices.Length:" + dazMesh.baseVertices.Length);
				Utility.debug("    materialNames:" + dazMesh.materialNames);
				Utility.debug("    materialNames.Length:" + dazMesh.materialNames.Length);
				foreach(var name in dazMesh.materialNames){
					Utility.debug("      name:" + name);
				}
				Utility.debug("    basePolyList:" + dazMesh.basePolyList);
				Utility.debug("    basePolyList.Length:" + dazMesh.basePolyList.Length);
				Utility.debug("    baseVerticesToUVVertices:" + dazMesh.baseVerticesToUVVertices);
				Utility.debug("    baseVerticesToUVVertices.Length:" + dazMesh.baseVerticesToUVVertices.Length);
				Utility.debug("    morphedUVVertices:" + dazMesh.morphedUVVertices);
				Utility.debug("    morphedUVVertices.Length:" + dazMesh.morphedUVVertices.Length);
				Utility.debug("    baseMesh:" + dazMesh.baseMesh);
				Utility.debug("    meshData:" + dazMesh.meshData);
				
			}
			Utility.debug("    hasClothSettings:" + hasClothSettings);
			if(hasClothSettings){
				ClothSimControl[] cscArray = this.cloth.GetComponentsInChildren<ClothSimControl>();
				foreach (ClothSimControl csc in cscArray){
					Utility.debug("      GeometryData:" + csc.clothSettings.GeometryData);
					
				}
			}
			var bones = cloth.GetComponentsInChildren<DAZBone>();
			Utility.debug("    bones:" + bones);
			foreach(var bone in bones){
				Utility.debug("      bone:" + bone);
			}

			DAZSkinWrapMaterialOptions wmo = cloth.GetComponentInChildren<DAZSkinWrapMaterialOptions>();
			Utility.debug("    wmo:" + wmo);
			if(wmo != null){
				Utility.debug("      customSimTextureClearButton:" + wmo.customSimTextureClearButton);
				if(wmo.customSimTextureClearButton != null){
//					wmo.customSimTextureClearButton.onClick.Invoke();
				}
				JSONStorableUrl jsu = wmo.GetComponent<JSONStorableUrl>();
				Utility.debug("      customSimTextureUrlJSON:" + jsu);
				wmo.SendMessage("SetCustomSimTextureToNull");
				wmo.BroadcastMessage("SetCustomSimTextureToNull");
				wmo.SendMessageUpwards("SetCustomSimTextureToNull");
			}
			var skinWraps = cloth.GetComponentsInChildren<DAZSkinWrap>();
			Utility.debug("    skinWraps:" + skinWraps);
			foreach(var skinWrap in skinWraps){
				Utility.debug("      skinWrap:" + skinWrap);
				Utility.debug("      skinWrap.simTextures:" + skinWrap.simTextures);
				
				foreach(var texture in skinWrap.simTextures){
					Utility.debug("        texture:" + texture);
					if(texture != null){
						Utility.debug("        texture.width:" + texture.width);
						Utility.debug("        texture.height:" + texture.height);
					}
				}
				Utility.debug("end");
			}

			var skinsV2 = cloth.GetComponentsInChildren<DAZSkinV2>();
			Utility.debug("    skinsV2:" + skinsV2);
			foreach(var skinV2 in skinsV2){
				Utility.debug("      skinV2:" + skinV2);
			}

			Utility.debug("  tags:" + cloth.tags);
			Utility.debug("  packageUid:" + cloth.packageUid);
			Utility.debug("  type:" + cloth.type);
		}
		
		public void WrapNormal(){
			wrapSwitcher = this.cloth.GetComponentInChildren<DAZSkinWrapSwitcher>();
			if(hasAltWrap && wrapSwitcher.currentWrapName != "Normal"){
				wrapSwitcher.SetCurrentWrapName("Normal");
			}
		}
		
		public void WrapAlter(bool isForce = false){
			wrapSwitcher = this.cloth.GetComponentInChildren<DAZSkinWrapSwitcher>();
			if(hasAltWrap){
				if(isForce || simEnabledOrig == false){
					int min = 0;
					int max = AltWrapNames.Count;
					int rand = UnityEngine.Random.Range(min, max);
					string altWrapName = AltWrapNames[rand];
					wrapSwitcher.SetCurrentWrapName(altWrapName);
				}
			}
		}
		
		public void Recover(){
			for(int i = 0; i < cssc.Count; i++){
				cssc[i].Recover();
			}
			WrapNormal();
		}
		
		public void Reset(){
			for(int i = 0; i < cssc.Count; i++){
				cssc[i].Reset();
			}
			WrapNormal();
		}
		
		public void Change(string type, float ratio){
			for(int i = 0; i < cssc.Count; i++){
				cssc[i].Change(type, ratio);
			}
		}
		
		public void UndressStop(){
			for(int i = 0; i < cssc.Count; i++){
				cssc[i].UndressStop();
			}
		}
		
		public void OnDestroy(){
			for(int i = 0; i < cssc.Count; i++){
				cssc[i].OnDestroy();
				cssc[i] = null;
			}
			AltWrapNames.Clear();
			AltWrapNames = null;
			wrapSwitcher = null;
			WrapNormal();
			cssc = null;
		}
	}
	
	public class ClothSimStateChanger
	{
		public ClothSimControl csc;
		public bool    backupBreakEnabled;
		public float   backupBreakThreshold;
		public float   backupFriction;
		public float   backupWeight;
		public float   backupGravityMultiplier;
		public float   backupStiffness;
		public bool    backupSimEnabled;
				
		public ClothSimStateChanger(ClothSimControl csc){
			this.csc = csc;
			backupBreakEnabled      = this.csc.clothSettings.BreakEnabled;
			backupBreakThreshold    = this.csc.clothSettings.BreakThreshold;
			backupFriction          = this.csc.clothSettings.Friction;
			backupWeight            = this.csc.clothSettings.Weight;
			backupGravityMultiplier = this.csc.clothSettings.GravityMultiplier;
			backupStiffness         = this.csc.clothSettings.Stiffness;
			backupSimEnabled        = this.csc.simEnabledJSON.val;
		}	

		public void Recover(){
			csc.clothSettings.BreakEnabled      = backupBreakEnabled;
			csc.clothSettings.BreakThreshold    = backupBreakThreshold;
			csc.clothSettings.Friction          = backupFriction;
			csc.clothSettings.Weight            = backupWeight;		
			csc.clothSettings.GravityMultiplier = backupGravityMultiplier;
			csc.clothSettings.Stiffness         = backupStiffness;
			csc.SetSimEnabled(backupSimEnabled);
		}
		
		public void Reset(){
			csc.Reset();
		}

		public void UndressStop(){
			csc.clothSettings.Friction          = 1.0f;
			csc.clothSettings.Weight            = 0.5f;
			csc.clothSettings.GravityMultiplier = 1.0f;
			csc.clothSettings.Stiffness = 1f;
			csc.clothSettings.BreakEnabled   = false;
			csc.clothSettings.BreakThreshold = 1f;
		}
		
		public void Change(string type, float ratio){
			if(type == "Undress"){
				csc.SetSimEnabled(true);
				csc.clothSettings.BreakEnabled   = true;
				csc.clothSettings.BreakThreshold = 0f;					
				Undress(ratio);
			}else if(type == "Break"){
				csc.SetSimEnabled(true);
				csc.clothSettings.BreakEnabled   = true;
				csc.clothSettings.BreakThreshold = 0.1f;
				csc.SetSimEnabled(true);
				csc.SyncAllowDetatch(true);
			}else if(type == "Break2"){
				csc.SetSimEnabled(true);
				csc.clothSettings.BreakEnabled   = true;
				csc.clothSettings.BreakThreshold = 0.0f;
				csc.SetSimEnabled(true);
				csc.SyncAllowDetatch(true);
			}else if(type == "Reverse"){
				Reverse(ratio);
				csc.clothSettings.Stiffness = 1f;
			}else if(type == "UndressAndReverse"){
				csc.SetSimEnabled(true);
				csc.clothSettings.BreakEnabled   = true;
				csc.clothSettings.BreakThreshold = 0f;			
				csc.clothSettings.Stiffness = 1f;
				Reverse(ratio);
			}else{
				Utility.debug("In ClothSimControl.Change: Unknown Command: " + type);
			}
		}
		
		public void Undress(float ratio){
			csc.clothSettings.Friction          = Mathf.Lerp(1.0f, 0.1f, ratio);
			csc.clothSettings.Weight            = Mathf.Lerp(0.5f, 2.0f, ratio);
			csc.clothSettings.GravityMultiplier = Mathf.Lerp(0.5f, 2.0f, ratio);			
		}
		
		public void Reverse(float ratio){
			csc.clothSettings.Friction          = Mathf.Lerp(1.0f,  0.1f, ratio);
			csc.clothSettings.Weight            = Mathf.Lerp(0.5f,  2.0f, ratio);
			csc.clothSettings.GravityMultiplier = Mathf.Lerp(1.0f, -2.0f, ratio);
		}
				
		public void OnDestroy(){
			csc = null;
		}
	}

}